#!/usr/bin/env node
'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var fs = require('fs');
var commander = require('commander');
var chalk = require('chalk');
var tslib = require('tslib');
var path = require('path');
require('ts-array-extensions');
var yaml = require('yaml');
var prettier = require('prettier');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var fs__default = /*#__PURE__*/_interopDefaultLegacy(fs);
var commander__default = /*#__PURE__*/_interopDefaultLegacy(commander);
var chalk__default = /*#__PURE__*/_interopDefaultLegacy(chalk);
var path__default = /*#__PURE__*/_interopDefaultLegacy(path);
var yaml__default = /*#__PURE__*/_interopDefaultLegacy(yaml);
var prettier__default = /*#__PURE__*/_interopDefaultLegacy(prettier);

/* eslint-disable no-console */
var o = process.stdout;
var e = process.stdout;
var progressLogger = function (start) {
    var last = 0;
    return function (m) {
        o.cursorTo(start);
        o.write(m);
        if (m.length < last) {
            o.write(' '.repeat(last - m.length));
        }
        last = m.length;
    };
};
var success = function () { return chalk__default['default'].green(' -> SUCCESS'); };
var warn = function (message) {
    return chalk__default['default'].yellow(" -> WARNING: " + message);
};
var progress = function (message) {
    return chalk__default['default'].blue(" -> " + message);
};
var processLogger = function () {
    var log = function (m) {
        o.write("\n" + m);
    };
    return {
        error: function (m) {
            e.write("\n" + chalk__default['default'].red(m));
        },
        log: log,
        create: function (m) {
            log(m);
            var l = progressLogger(m.length);
            l(progress('...'));
            return l;
        }
    };
};

var mapProperties$1 = function (properties) {
    var r = {};
    Object.keys(properties).forEach(function (key) {
        var property = properties[key];
        if ('format' in property && property.format === 'binary') {
            return;
        }
        // eslint-disable-next-line @typescript-eslint/no-use-before-define
        r[key] = convertSchemaToJsonSchema$1(properties[key]);
    });
    return r;
};
var convertSchemaToJsonSchema$1 = function (schema) {
    if ('$ref' in schema) {
        return {
            $ref: schema.$ref.replace('components/schemas', 'definitions')
        };
    }
    return tslib.__assign(tslib.__assign(tslib.__assign(tslib.__assign(tslib.__assign(tslib.__assign({}, schema), ('properties' in schema && schema.properties
        ? { properties: mapProperties$1(schema.properties) }
        : undefined)), (schema.allOf
        ? { allOf: schema.allOf.map(convertSchemaToJsonSchema$1) }
        : undefined)), (schema.anyOf
        ? { anyOf: schema.anyOf.map(convertSchemaToJsonSchema$1) }
        : undefined)), (schema.oneOf
        ? { oneOf: schema.oneOf.map(convertSchemaToJsonSchema$1) }
        : undefined)), ('items' in schema && schema.items
        ? { items: convertSchemaToJsonSchema$1(schema.items) }
        : undefined));
};

var dereference$1 = function (references, $ref, prefix) {
    var current = {
        $ref: $ref
    };
    while (current && '$ref' in current) {
        if (!current.$ref.startsWith(prefix)) {
            throw new Error("Invalid $ref " + $ref);
        }
        var name = current.$ref.substring(current.$ref.lastIndexOf('/') + 1);
        current = references === null || references === void 0 ? void 0 : references[name];
    }
    if (!current)
        throw new Error("Failed to deference " + $ref);
    return current;
};

var dereferenceParameter$1 = function (reference, doc) {
    var _a;
    var $ref = reference.$ref;
    return dereference$1((_a = doc.components) === null || _a === void 0 ? void 0 : _a.parameters, $ref, '#/components/parameters/');
};

var dereferenceRequestBodyObject = function (reference, doc) {
    var _a;
    var $ref = reference.$ref;
    return dereference$1((_a = doc.components) === null || _a === void 0 ? void 0 : _a.requestBodies, $ref, '#/components/requestBodies/');
};

var dereferenceResponseObject$1 = function (reference, doc) {
    var _a;
    var $ref = reference.$ref;
    return dereference$1((_a = doc.components) === null || _a === void 0 ? void 0 : _a.responses, $ref, '#/components/responses/');
};

var dereferenceSchemaObject = function (reference, doc) {
    var _a;
    var $ref = reference.$ref;
    return dereference$1((_a = doc.components) === null || _a === void 0 ? void 0 : _a.schemas, $ref, '#/components/schemas/');
};

function isReferenceObject$1(object) {
    if (typeof object !== 'object') {
        return false;
    }
    if (object === null) {
        return false;
    }
    return '$ref' in object;
}

var methodsOfPathItemObject$1 = function (pathItemObject) {
    return Object.keys(pathItemObject).compactMap(function (method) {
        if (method === '$ref' ||
            method === 'parameters' ||
            method === 'summary' ||
            method === 'description' ||
            method === 'servers') {
            return undefined;
        }
        var op = pathItemObject[method];
        if (!op) {
            return undefined;
        }
        return [method, op];
    });
};

var mapOperations$1 = function (document) {
    return Object.keys(document.paths).flatMap(function (path) {
        var _a;
        if (typeof path !== 'string')
            return [];
        var pathObject = document.paths[path];
        var pathParameters = (_a = pathObject.parameters) !== null && _a !== void 0 ? _a : [];
        return methodsOfPathItemObject$1(pathObject).map(function (_a) {
            var _b = tslib.__read(_a, 2), method = _b[0], operationObject = _b[1];
            if (!('operationId' in operationObject) ||
                !operationObject.operationId) {
                throw new Error("OperationId required for " + method + " " + path);
            }
            var operationId = operationObject.operationId, _c = operationObject.parameters, parameters = _c === void 0 ? [] : _c, responses = operationObject.responses, requestBody = operationObject.requestBody;
            var inlinedParameters = pathParameters
                .concat(parameters)
                .map(function (p) {
                return isReferenceObject$1(p) ? dereferenceParameter$1(p, document) : p;
            });
            var inlinedRequestBody = requestBody && isReferenceObject$1(requestBody)
                ? dereferenceRequestBodyObject(requestBody, document)
                : requestBody;
            var inlinedResponses = Object.keys(responses).compactMap(function (statusCode) {
                var response = responses === null || responses === void 0 ? void 0 : responses[statusCode];
                if (!response)
                    return undefined;
                return {
                    statusCode: statusCode.toString(),
                    response: isReferenceObject$1(response)
                        ? dereferenceResponseObject$1(response, document)
                        : response
                };
            });
            return {
                path: path,
                method: method,
                operationId: operationId,
                operation: operationObject,
                parameters: inlinedParameters !== null && inlinedParameters !== void 0 ? inlinedParameters : [],
                requestBody: inlinedRequestBody,
                responses: inlinedResponses
            };
        });
    });
};

var typeDefForEnum$2 = function (members) {
    return members
        .map(function (e) {
        if (typeof e === 'string') {
            return "\"" + e + "\"";
        }
        if (e === null) {
            return 'null';
        }
        if (typeof e === 'number') {
            return "" + e;
        }
        if (typeof e === 'object') {
            return JSON.stringify(e);
        }
        throw new Error("Unexpected enum member");
    })
        .join(' | ');
};

var typeDefForObject$2 = function (objectSchema, recursiveLookup) {
    var required = objectSchema.required, properties = objectSchema.properties;
    if (!properties)
        return 'unknown';
    var requiredProperties = Array.isArray(required) ? required : [];
    var propertyList = Object.keys(properties).compactMap(function (propertyName) {
        if (typeof propertyName !== 'string') {
            return undefined;
        }
        var propertyObject = properties[propertyName];
        var isRequired = requiredProperties.includes(propertyName);
        var propertyType = recursiveLookup(propertyObject);
        return "" + propertyName + (isRequired ? '' : '?') + ": " + propertyType;
    });
    return "{ " + propertyList.join('; ') + " }";
};

var typeDefForReference$1 = function ($ref) {
    var name = $ref.substring($ref.lastIndexOf('/') + 1);
    return name;
};

var typeDefForSchema$2 = function (schema, binaryType) {
    if (binaryType === void 0) { binaryType = 'Buffer'; }
    if ('$ref' in schema) {
        return typeDefForReference$1(schema.$ref);
    }
    var enumProp = schema.enum, allOf = schema.allOf, oneOf = schema.oneOf, anyOf = schema.anyOf;
    if (enumProp) {
        return typeDefForEnum$2(enumProp);
    }
    var base = schema.type === 'object'
        ? typeDefForObject$2(schema, typeDefForSchema$2)
        : null;
    if (allOf) {
        return allOf
            .map(function (m) { return typeDefForSchema$2(m); })
            .concat(base ? [base] : [])
            .join(' & ');
    }
    var unionOf = (oneOf !== null && oneOf !== void 0 ? oneOf : []).concat(anyOf !== null && anyOf !== void 0 ? anyOf : []);
    if (unionOf.length > 0) {
        var union = unionOf.map(function (m) { return typeDefForSchema$2(m); }).join(' | ');
        return base ? base + " & (" + union + ")" : union;
    }
    if (schema.type === 'object') {
        return base !== null && base !== void 0 ? base : 'unknown';
    }
    if (schema.type === 'string') {
        if (schema.format === 'binary') {
            return binaryType;
        }
        return 'string';
    }
    if (schema.type === 'boolean' || schema.type === 'number') {
        return schema.type;
    }
    if (schema.type === 'integer') {
        return 'number';
    }
    if (schema.type === 'array') {
        var items = schema.items;
        if (!items) {
            throw new Error("Array nodes must have an items property.");
        }
        if (Array.isArray(items)) {
            throw new Error("Array items property must be singular.");
        }
        return "Array<" + typeDefForSchema$2(items) + ">";
    }
    throw new Error("Unrecognized type in schema " + JSON.stringify(schema));
};

var fieldTemplate = function (name, required, type) {
    return "" + name + (required ? '' : '?') + ": " + type;
};

var objectTemplate = function (fields) {
    return fields.none()
        ? 'unknown'
        : "{" + fields
            .map(function (f) { return fieldTemplate(f.name, f.required, f.type); })
            .join(';') + "}";
};

var clientMethodTemplate = function (_a) {
    var httpMethod = _a.httpMethod, methodName = _a.methodName, openApiPath = _a.openApiPath, pathParams = _a.pathParams, queryParams = _a.queryParams, headerParams = _a.headerParams, body = _a.body, responses = _a.responses;
    var bodyArg = (function () {
        if (!body)
            return undefined;
        var typeDef = body.type === 'json'
            ? body.jsonType
            : objectTemplate(body.fields.map(function (f) { return ({
                name: f.name,
                required: f.required,
                type: f.binary ? 'Blob' : 'string'
            }); }));
        return {
            name: 'body',
            required: body.required,
            type: typeDef
        };
    })();
    var functionArguments = queryParams
        .concat(headerParams)
        .map(function (_a) {
        var name = _a.name, required = _a.required;
        return ({ name: name, required: required, type: 'string' });
    })
        .concat(pathParams.map(function (name) { return ({ name: name, required: true, type: 'string' }); }))
        .concat(bodyArg ? [bodyArg] : []);
    var functionArgumentSignature = functionArguments.any()
        ? "args: " + objectTemplate(functionArguments)
        : '';
    var responseType = responses
        .map(function (r) {
        var statusCodeType = r.statusCode === 'default' ? 'number' : r.statusCode;
        var dataType = (function () {
            if (r.type === 'binary')
                return 'Blob';
            if (r.type === 'json')
                return r.jsonType;
            return 'undefined';
        })();
        return "ResponseWithData<" + statusCodeType + ", " + dataType + ">";
    })
        .join(' | ');
    var paramNames = headerParams
        .concat(queryParams)
        .map(function (_a) {
        var name = _a.name;
        return name;
    })
        .concat(pathParams)
        .concat(bodyArg ? [bodyArg.name] : []);
    var decomposeParameters = paramNames.any()
        ? "const {" + paramNames.join(', ') + "} = args;"
        : '';
    var composeQuery = queryParams.any()
        ? "const query = qs.stringify({ " + queryParams
            .map(function (q) { return q.name; })
            .join(', ') + " });"
        : '';
    var composeFormData = (body === null || body === void 0 ? void 0 : body.type) === 'form'
        ? "const formData = new FormData();\n      " + body.fields
            .map(function (f) { return "formData.append('" + f.name + "', body." + f.name + ");"; })
            .join('\n') + "\n    "
        : '';
    var url = "`${this.baseUrl}" + openApiPath.replace(/\{(?:.*?)\}/g, function (x) { return "$" + x; }) + (queryParams.any() ? '?${query}' : '') + "`";
    var headers = headerParams
        .map(function (h) { return h.name; })
        .concat((body === null || body === void 0 ? void 0 : body.type) === 'json' ? ["'Content-Type': 'application/json'"] : []);
    var fetchOpts = "{\n      " + tslib.__spreadArray(tslib.__spreadArray(tslib.__spreadArray([
        'method'
    ], tslib.__read((headers.any() ? ["headers: {" + headers.join(',') + "}"] : []))), tslib.__read(((body === null || body === void 0 ? void 0 : body.type) === 'json' ? ['body: JSON.stringify(body)'] : []))), tslib.__read(((body === null || body === void 0 ? void 0 : body.type) === 'form' ? ['body: formData'] : []))).join(',\n') + "\n    }";
    var responseSwitch = "\n    switch ($status) {\n      " + responses
        .map(function (r) {
        var bodyExpression = (function () {
            if (r.type === 'json') {
                return "(await response.json()) as " + r.jsonType;
            }
            if (r.type === 'binary')
                return 'await response.blob()';
            return 'undefined';
        })();
        return "\n              " + (r.statusCode === 'default' ? "default" : "case " + r.statusCode) + ": return {\n                status: $status,\n                body: " + bodyExpression + "\n              }\n            ";
    })
        .join('\n') + "\n\n      " + (responses.every(function (r) { return r.statusCode !== 'default'; })
        ? "\n          default:\n            throw new Error(" +
            '`Unexpected status ${$status}`' +
            ")\n        "
        : '') + "\n    }\n  ";
    return "public async " + methodName + "(" + functionArgumentSignature + "): \n    Promise<" + responseType + "> {\n\n    " + decomposeParameters + "\n\n    " + composeQuery + "\n\n    " + composeFormData + "\n\n    const method = '" + httpMethod.toUpperCase() + "',\n    const url = " + url + ";\n\n    const response = await fetch(url, " + fetchOpts + ");\n\n    const { status: $status } = response;\n\n    this.logger?.('REST API Call', { method, url, status: $status });\n\n    " + responseSwitch + "\n  }";
};

var initUpper = function (value) {
    return value.replace(/(?:^|\s)[a-z]/g, function (m) { return m.toUpperCase(); });
};

var parametersTypeNameTemplate = function (operationId, iin) {
    return initUpper(operationId) + "Request" + initUpper(iin === 'formData' ? 'body' : iin);
};

var requestHandlerTemplate = function (_a) {
    var typeName = _a.typeName, _b = _a.pathParamsType, pathParamsType = _b === void 0 ? 'unknown' : _b, _c = _a.responseBodyType, responseBodyType = _c === void 0 ? 'unknown' : _c, _d = _a.bodyParamType, bodyParamType = _d === void 0 ? 'unknown' : _d, _e = _a.queryParamsType, queryParamsType = _e === void 0 ? 'ParsedQs' : _e, _f = _a.statusCodeType, statusCodeType = _f === void 0 ? 'number' : _f;
    return "\n    export type " + typeName + " = RequestHandler<\n      " + pathParamsType + ",\n      " + responseBodyType + ",\n      " + bodyParamType + ",\n      " + queryParamsType + ",\n      Record<string, any>,\n      " + statusCodeType + "\n    >";
};

var requestHandlersTemplate = function (handlers) {
    return "\n    " + handlers.map(requestHandlerTemplate).join(';\n') + "\n\n    export type RequestHandlers = {\n      " + handlers
        .map(function (_a) {
        var methodName = _a.methodName, typeName = _a.typeName;
        return "\n        " + methodName + ": " + typeName + "\n      ";
    })
        .join(';\n') + "\n    }";
};

var initLower = function (value) {
    return value.replace(/(?:^|\s)[A-Z]/g, function (m) { return m.toLowerCase(); });
};

var requestValidatorFunctionNameTemplate = function (operationId) {
    return initLower(operationId) + "Validator";
};

/* eslint-disable no-template-curly-in-string */
var requestValidatorTemplate = function (_a) {
    var functionName = _a.functionName, parameterTypes = _a.parameterTypes;
    return ("\n    const " + functionName + " = (options?: ValidationOptions): RequestHandler => {\n      " + parameterTypes
        .map(function (_a) {
        var typeName = _a.typeName, requestProperty = _a.requestProperty;
        return "\n        const " + requestProperty + " = ajv.getSchema('#/definitions/" + typeName + "')!;\n      ";
    })
        .join('\n') + "\n\n      return (req, res, next) => {\n        if ([\n          " + parameterTypes
        .map(function (_a) {
        var requestProperty = _a.requestProperty;
        return "\n            " + requestProperty + "(req." + requestProperty + ")\n          ";
    })
        .join(',\n') + "\n        ].every(r => r)) {\n          return next();\n        }\n\n        const fields = ([\n          " + parameterTypes
        .map(function (_a) {
        var requestProperty = _a.requestProperty;
        return "\n            [" + requestProperty + ", '" + requestProperty + "']\n          ";
    })
        .join(',\n') + "\n        ] as const).flatMap(([validator, path]) =>\n          validator.errors?.map(e => ({\n            path: " +
        '`${path}${e.dataPath}`,' +
        "\n            message: e.message\n          }))\n        ).compact();\n\n        options?.logger?.(req)('Request validation failed', { fields });\n\n        res.status(400).send({\n          type: 'REQUEST_VALIDATION_FAILED',\n          fields\n        });\n      };\n    };\n  ");
};

var responseBodyTypeNameTemplate = function (operationId, statusCode) { return "" + initUpper(operationId) + statusCode + "ResponseBody"; };

var routerTemplate = function (_a) {
    var serviceName = _a.serviceName, routes = _a.routes;
    var argumentList = routes
        .map(function (_a) {
        var handlerInstanceName = _a.handlerInstanceName;
        return handlerInstanceName;
    })
        .join(', ');
    return "\n    export const " + initLower(serviceName) + "Router\n      = ({ " + argumentList + " }: RequestHandlers, options?: ValidationOptions): Router => {\n      return Router()\n        " + routes
        .map(function (t) {
        var path = t.openApiPath.replace(/\{(?:.*?)\}/g, function (x) { return ":" + x.substr(1, x.length - 2); });
        var args = tslib.__spreadArray(tslib.__spreadArray([
            "'" + path + "'"
        ], tslib.__read((t.validatorInstanceName
            ? [t.validatorInstanceName + "(options)"]
            : []))), [
            t.handlerInstanceName
        ]);
        return "." + t.httpMethod + "(" + args.join(', ') + ")";
    })
        .join('\n') + "\n    }\n  ";
};

var serviceNameTemplate = function (title) {
    return title.replace(/\s+.{1}/g, function (v) { return v.slice(-1).toUpperCase(); });
};

var fromV3$7 = function (document, log) {
    return mapOperations$1(document).map(function (_a) {
        var method = _a.method, parameters = _a.parameters, path = _a.path, operationId = _a.operationId, responses = _a.responses, requestBody = _a.requestBody;
        log === null || log === void 0 ? void 0 : log(progress("Adding " + method + " " + path));
        var body = (function () {
            var _a;
            if (!requestBody)
                return null;
            if ('application/json' in requestBody.content) {
                var schema = requestBody.content['application/json'].schema;
                return {
                    type: 'json',
                    required: (_a = requestBody.required) !== null && _a !== void 0 ? _a : false,
                    jsonType: schema ? typeDefForSchema$2(schema) : 'unknown'
                };
            }
            if ('multipart/form-data' in requestBody.content) {
                var schema = requestBody.content['multipart/form-data'].schema;
                if (!schema)
                    return null;
                var inlinedSchema_1 = isReferenceObject$1(schema)
                    ? dereferenceSchemaObject(schema, document)
                    : schema;
                var fields = inlinedSchema_1.properties
                    ? Object.keys(inlinedSchema_1.properties).map(function (k) {
                        var _a, _b, _c;
                        var property = (_a = inlinedSchema_1.properties) === null || _a === void 0 ? void 0 : _a[k];
                        var inlinedProperty = isReferenceObject$1(property)
                            ? dereferenceSchemaObject(property, document)
                            : property;
                        return {
                            name: k.toString(),
                            required: (_c = (_b = inlinedSchema_1.required) === null || _b === void 0 ? void 0 : _b.includes(k.toString())) !== null && _c !== void 0 ? _c : false,
                            binary: (inlinedProperty === null || inlinedProperty === void 0 ? void 0 : inlinedProperty.format) === 'binary'
                        };
                    })
                    : [];
                return {
                    type: 'form',
                    required: requestBody.required,
                    fields: fields
                };
            }
            return null;
        })();
        return clientMethodTemplate({
            httpMethod: method,
            methodName: operationId,
            openApiPath: path,
            pathParams: parameters.filter(function (p) { return p.in === 'path'; }).map(function (p) { return p.name; }),
            queryParams: parameters
                .filter(function (p) { return p.in === 'query'; })
                .map(function (p) {
                var _a;
                return ({
                    name: p.name,
                    required: (_a = p.required) !== null && _a !== void 0 ? _a : false
                });
            }),
            headerParams: parameters
                .filter(function (p) { return p.in === 'header'; })
                .map(function (p) {
                var _a;
                return ({
                    name: p.name,
                    required: (_a = p.required) !== null && _a !== void 0 ? _a : false
                });
            }),
            body: body,
            responses: responses.map(function (_a) {
                var statusCode = _a.statusCode, response = _a.response;
                if (!response.content || Object.keys(response.content).length === 0) {
                    return {
                        statusCode: statusCode,
                        type: 'none'
                    };
                }
                if (response.content && 'application/json' in response.content) {
                    var schema = response.content['application/json'].schema;
                    return {
                        statusCode: statusCode,
                        type: 'json',
                        jsonType: schema ? typeDefForSchema$2(schema) : 'unknown'
                    };
                }
                return {
                    statusCode: statusCode,
                    type: 'binary'
                };
            })
        });
    });
};

var mapProperties = function (properties) {
    var r = {};
    Object.keys(properties).forEach(function (key) {
        // eslint-disable-next-line @typescript-eslint/no-use-before-define
        r[key] = convertSchemaToJsonSchema(properties[key]);
    });
    return r;
};
var convertSchemaToJsonSchema = function (schema) {
    /**
     * This currently does nothing, but is a placeholder for removing
     * unsupported features from Open Api
     */
    var properties = schema.properties, rest = tslib.__rest(schema, ["properties"]);
    return tslib.__assign(tslib.__assign({}, rest), (properties ? { properties: mapProperties(properties) } : undefined));
};

var dereference = function (references, $ref, prefix) {
    if (!$ref.startsWith(prefix)) {
        throw new Error("Invalid $ref " + $ref);
    }
    var name = $ref.substring($ref.lastIndexOf('/') + 1);
    var maybe = references === null || references === void 0 ? void 0 : references[name];
    if (!maybe)
        throw new Error("Failed to deference " + $ref);
    return maybe;
};

var dereferenceParameter = function (reference, doc) {
    var $ref = reference.$ref;
    return dereference(doc.parameters, $ref, '#/parameters/');
};

var dereferenceResponseObject = function (reference, doc) {
    var $ref = reference.$ref;
    return dereference(doc.responses, $ref, '#/responses/');
};

function isReferenceObject(object) {
    if (typeof object !== 'object') {
        return false;
    }
    if (object === null) {
        return false;
    }
    return '$ref' in object;
}

var methodsOfPathItemObject = function (pathItemObject) {
    return Object.keys(pathItemObject).compactMap(function (method) {
        if (method === '$ref' || method === 'parameters') {
            return undefined;
        }
        var op = pathItemObject[method];
        if (!op) {
            return undefined;
        }
        return [method, op];
    });
};

var mapOperations = function (document) {
    return Object.keys(document.paths).flatMap(function (path) {
        var _a;
        if (typeof path !== 'string')
            return [];
        var pathObject = document.paths[path];
        var pathParameters = (_a = pathObject.parameters) !== null && _a !== void 0 ? _a : [];
        return methodsOfPathItemObject(pathObject).map(function (_a) {
            var _b = tslib.__read(_a, 2), method = _b[0], operationObject = _b[1];
            if (!('operationId' in operationObject) ||
                !operationObject.operationId) {
                throw new Error("OperationId required for " + method + " " + path);
            }
            var operationId = operationObject.operationId, _c = operationObject.parameters, parameters = _c === void 0 ? [] : _c, responses = operationObject.responses;
            var inlinedParameters = pathParameters
                .concat(parameters)
                .map(function (p) {
                return isReferenceObject(p) ? dereferenceParameter(p, document) : p;
            });
            var inlinedResponses = Object.keys(responses).compactMap(function (statusCode) {
                var response = responses[statusCode];
                if (!response)
                    return undefined;
                return {
                    statusCode: statusCode.toString(),
                    response: isReferenceObject(response)
                        ? dereferenceResponseObject(response, document)
                        : response
                };
            });
            return {
                path: path,
                method: method,
                operationId: operationId,
                operation: operationObject,
                parameters: inlinedParameters !== null && inlinedParameters !== void 0 ? inlinedParameters : [],
                responses: inlinedResponses
            };
        });
    });
};

var typeDefForEnum$1 = function (members) {
    return members
        .map(function (e) {
        if (typeof e === 'string') {
            return "\"" + e + "\"";
        }
        if (e === null) {
            return 'null';
        }
        if (typeof e === 'number') {
            return "" + e;
        }
        if (typeof e === 'object') {
            return JSON.stringify(e);
        }
        throw new Error("Unexpected enum member");
    })
        .join(' | ');
};

var typeDefForObject$1 = function (objectSchema, recursiveLookup) {
    var required = objectSchema.required, properties = objectSchema.properties;
    if (!properties)
        return 'unknown';
    var requiredProperties = Array.isArray(required) ? required : [];
    var propertyList = Object.keys(properties).compactMap(function (propertyName) {
        if (typeof propertyName !== 'string') {
            return undefined;
        }
        var propertyObject = properties[propertyName];
        var isRequired = requiredProperties.includes(propertyName);
        var propertyType = recursiveLookup(propertyObject);
        return "" + propertyName + (isRequired ? '' : '?') + ": " + propertyType;
    });
    return "{ " + propertyList.join('; ') + " }";
};

var typeDefForSchema$1 = function (schema) {
    var _a;
    if (typeof schema === 'boolean') {
        throw new Error("Schema nodes must be an object.");
    }
    if ('$ref' in schema && schema.$ref) {
        return typeDefForReference$1(schema.$ref);
    }
    if (!('type' in schema)) {
        throw new Error("Schema nodes must have a type. " + JSON.stringify(schema));
    }
    var type = schema.type, enumProp = schema.enum, allOf = schema.allOf;
    if (enumProp) {
        return typeDefForEnum$1(enumProp);
    }
    if (type === 'object') {
        var def = typeDefForObject$1(schema, typeDefForSchema$1);
        if (allOf) {
            return allOf
                .map(function (m) { return typeDefForSchema$1(m); })
                .concat(def)
                .join(' & ');
        }
        return def;
    }
    if (allOf) {
        return allOf
            .map(function (m) { return typeDefForSchema$1(m); })
            .join(' & ');
    }
    if (type === 'boolean' ||
        type === 'null' ||
        type === 'number' ||
        type === 'string') {
        return type;
    }
    if (type === 'integer') {
        return 'number';
    }
    if (type === 'array') {
        var items = schema.items;
        if (!items) {
            throw new Error("Array nodes must have an items property.");
        }
        if (Array.isArray(items)) {
            throw new Error("Array items property must be singular.");
        }
        return "Array<" + typeDefForSchema$1(items) + ">";
    }
    throw new Error("Unrecognized type " + ((_a = type === null || type === void 0 ? void 0 : type.toString()) !== null && _a !== void 0 ? _a : ''));
};

var fromV2$7 = function (document, log) {
    return mapOperations(document).map(function (_a) {
        var _b;
        var method = _a.method, parameters = _a.parameters, path = _a.path, operationId = _a.operationId, responses = _a.responses;
        log === null || log === void 0 ? void 0 : log(progress("Adding " + method + " " + path));
        var jsonBody = function () {
            var _a;
            var bodyArg = parameters.filter(function (p) { return p.in === 'body'; }).first();
            if (!bodyArg)
                return null;
            return {
                type: 'json',
                jsonType: typeDefForSchema$1(bodyArg.schema),
                required: (_a = bodyArg.required) !== null && _a !== void 0 ? _a : false
            };
        };
        var formDataBody = function () {
            var bodyArg = parameters.filter(function (p) { return p.in === 'formData'; });
            if (bodyArg.length === 0)
                return null;
            return {
                type: 'form',
                fields: bodyArg.map(function (p) {
                    var _a;
                    return ({
                        name: p.name,
                        required: (_a = p.required) !== null && _a !== void 0 ? _a : false,
                        binary: p.type === 'file'
                    });
                }),
                required: true
            };
        };
        return clientMethodTemplate({
            httpMethod: method,
            methodName: operationId,
            openApiPath: path,
            pathParams: parameters.filter(function (p) { return p.in === 'path'; }).map(function (p) { return p.name; }),
            queryParams: parameters
                .filter(function (p) { return p.in === 'query'; })
                .map(function (p) {
                var _a;
                return ({
                    name: p.name,
                    required: (_a = p.required) !== null && _a !== void 0 ? _a : false
                });
            }),
            headerParams: parameters
                .filter(function (p) { return p.in === 'header'; })
                .map(function (p) {
                var _a;
                return ({
                    name: p.name,
                    required: (_a = p.required) !== null && _a !== void 0 ? _a : false
                });
            }),
            body: (_b = jsonBody()) !== null && _b !== void 0 ? _b : formDataBody(),
            responses: responses.map(function (_a) {
                var _b;
                var statusCode = _a.statusCode, response = _a.response;
                if (((_b = response.schema) === null || _b === void 0 ? void 0 : _b.type) === 'file') {
                    return {
                        statusCode: statusCode,
                        type: 'binary'
                    };
                }
                var schema = response.schema;
                if (schema) {
                    return {
                        statusCode: statusCode,
                        type: 'json',
                        jsonType: schema ? typeDefForSchema$1(schema) : 'unknown'
                    };
                }
                return {
                    statusCode: statusCode,
                    type: 'none'
                };
            })
        });
    });
};

var generateMethods = function (openApiDocument, log) {
    return 'swagger' in openApiDocument
        ? fromV2$7(openApiDocument, log)
        : fromV3$7(openApiDocument, log);
};

var fromV2$6 = function (document, log) {
    var _a;
    return Object.keys((_a = document.definitions) !== null && _a !== void 0 ? _a : {}).compactMap(function (key) {
        var _a;
        var value = (_a = document.definitions) === null || _a === void 0 ? void 0 : _a[key];
        if (!value) {
            return undefined;
        }
        log === null || log === void 0 ? void 0 : log(progress("adding " + key));
        var typeDef = typeDefForSchema$1(value);
        return "export type " + key + " = " + typeDef + ";";
    });
};

var fromV3$6 = function (document, log) {
    var _a, _b;
    return Object.keys((_b = (_a = document.components) === null || _a === void 0 ? void 0 : _a.schemas) !== null && _b !== void 0 ? _b : {}).compactMap(function (key) {
        var _a, _b;
        var value = (_b = (_a = document.components) === null || _a === void 0 ? void 0 : _a.schemas) === null || _b === void 0 ? void 0 : _b[key];
        if (!value) {
            return undefined;
        }
        log === null || log === void 0 ? void 0 : log(progress("adding " + key));
        var typeDef = typeDefForSchema$2(value);
        return "export type " + key + " = " + typeDef + ";";
    });
};

var generateTypes = function (openApiDocument, log) {
    return 'swagger' in openApiDocument
        ? fromV2$6(openApiDocument, log)
        : fromV3$6(openApiDocument, log);
};

var generateClient = function (_a) {
    var logger = _a.logger, openApiDocument = _a.openApiDocument, serviceName = _a.serviceName;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating client');
    var types = generateTypes(openApiDocument, log);
    var methods = generateMethods(openApiDocument, log);
    var code = "\n    import qs from 'qs';\n    import { fetch } from 'cross-fetch';\n\n    " + types.join('\n\n') + "\n\n    type ResponseWithData<TStatus, TData> = {\n      status: TStatus;\n      body: TData;\n    };\n      \n    type LogFn = (message: string, data: Record<string, unknown>) => void;\n\n    export class " + serviceName + "Client {\n      public constructor(\n        public readonly baseUrl: string,\n        private readonly logger?: LogFn\n      ) {}\n      \n      " + methods.join('\n\n') + "\n    }\n  ";
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var fromV2$5 = function (document, log) {
    return mapOperations(document).flatMap(function (_a) {
        var operationId = _a.operationId, responses = _a.responses;
        return responses.compactMap(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var schema = response.schema;
            if (!schema) {
                return undefined;
            }
            if (schema.type !== 'file') {
                return undefined;
            }
            var typeName = responseBodyTypeNameTemplate(operationId, statusCode);
            log === null || log === void 0 ? void 0 : log(progress("adding " + typeName));
            return "export type " + typeName + " = Buffer;";
        });
    });
};

var fromV3$5 = function (document, log) {
    return mapOperations$1(document).flatMap(function (_a) {
        var operationId = _a.operationId, responses = _a.responses;
        return responses.compactMap(function (_a) {
            var _b;
            var statusCode = _a.statusCode, response = _a.response;
            if ((_b = response.content) === null || _b === void 0 ? void 0 : _b['application/json']) {
                return undefined;
            }
            var typeName = responseBodyTypeNameTemplate(operationId, statusCode);
            log === null || log === void 0 ? void 0 : log(progress("adding " + typeName));
            return "export type " + typeName + " = Buffer;";
        });
    });
};

var typeDefForObject = function (objectSchema, recursiveLookup) {
    var required = objectSchema.required, properties = objectSchema.properties;
    if (!properties)
        return 'unknown';
    var requiredProperties = Array.isArray(required) ? required : [];
    var propertyList = Object.keys(properties).compactMap(function (propertyName) {
        if (typeof propertyName !== 'string') {
            return undefined;
        }
        var propertyObject = properties[propertyName];
        var isRequired = requiredProperties.includes(propertyName);
        var propertyType = recursiveLookup(propertyObject);
        return "" + propertyName + (isRequired ? '' : '?') + ": " + propertyType;
    });
    return "{ " + propertyList.join('; ') + " }";
};

var typeDefForEnum = function (members, onwardLookup) {
    return members
        .map(function (e) {
        if (typeof e === 'string') {
            return "\"" + e + "\"";
        }
        if (e === null) {
            return 'null';
        }
        if (typeof e === 'object') {
            if (Array.isArray(e)) {
                throw new Error("Don't know what to do with an array-like enum member");
            }
            return onwardLookup(e);
        }
        return e.toString();
    })
        .join(' | ');
};

var typeDefForReference = function ($ref) {
    var name = $ref.substring($ref.lastIndexOf('/') + 1);
    return name;
};

var typeDefForSchema = function (schema) {
    if (typeof schema === 'boolean') {
        throw new Error("Schema nodes must be an object.");
    }
    if (Array.isArray(schema.type)) {
        throw new Error("Array schema types are not supported (until we understand what they're for).");
    }
    if ('$ref' in schema && schema.$ref) {
        return typeDefForReference(schema.$ref);
    }
    var enumProp = schema.enum, allOf = schema.allOf, oneOf = schema.oneOf, anyOf = schema.anyOf;
    if (enumProp) {
        return typeDefForEnum(enumProp, typeDefForSchema);
    }
    var base = schema.type === 'object'
        ? typeDefForObject(schema, typeDefForSchema)
        : null;
    if (allOf) {
        return allOf
            .map(function (m) { return typeDefForSchema(m); })
            .concat(base ? [base] : [])
            .join(' & ');
    }
    var unionOf = (oneOf !== null && oneOf !== void 0 ? oneOf : []).concat(anyOf !== null && anyOf !== void 0 ? anyOf : []);
    if (unionOf.length > 0) {
        var union = unionOf.map(function (m) { return typeDefForSchema(m); }).join(' | ');
        return base ? base + " & (" + union + ")" : union;
    }
    if (schema.type === 'object') {
        return base !== null && base !== void 0 ? base : 'unknown';
    }
    if (schema.type === 'boolean' ||
        schema.type === 'null' ||
        schema.type === 'number' ||
        schema.type === 'string') {
        return schema.type;
    }
    if (schema.type === 'integer') {
        return 'number';
    }
    if (schema.type === 'array') {
        var items = schema.items;
        if (!items) {
            throw new Error("Array nodes must have an items property.");
        }
        if (Array.isArray(items)) {
            throw new Error("Array items property must be singular.");
        }
        return "Array<" + typeDefForSchema(items) + ">";
    }
    throw new Error("Unrecognized type in schema " + JSON.stringify(schema));
};

var fromJsonSchema = function (jsonSchema, log) {
    var definitions = jsonSchema.definitions;
    if (!definitions)
        return [];
    return Object.keys(definitions).compactMap(function (typeName) {
        if (typeof typeName !== 'string')
            return undefined;
        var definition = definitions[typeName];
        log === null || log === void 0 ? void 0 : log(progress(typeName));
        return "export type " + typeName + " = " + typeDefForSchema(definition) + ";";
    });
};

var generateServerTypes = function (openApiDocument, jsonSchema, log) {
    log === null || log === void 0 ? void 0 : log(progress('Generating typescript types'));
    var types = tslib.__spreadArray(tslib.__spreadArray([], tslib.__read(fromJsonSchema(jsonSchema, log))), tslib.__read(('swagger' in openApiDocument
        ? fromV2$5(openApiDocument, log)
        : fromV3$5(openApiDocument, log))));
    log === null || log === void 0 ? void 0 : log(success());
    return types.join('\n\n');
};

var fromV2$4 = function (document, log) {
    var handlers = mapOperations(document).map(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters, responses = _a.responses;
        var responseBodyType = responses
            .compactMap(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var schema = response.schema;
            if (!schema) {
                return undefined;
            }
            return responseBodyTypeNameTemplate(operationId, statusCode);
        })
            .join(' | ') || 'unknown';
        var statusCodes = responses
            .map(function (_a) {
            var statusCode = _a.statusCode;
            return statusCode === 'default' ? 'number' : statusCode;
        })
            .join(' | ');
        var bodyParamType = parameters.some(function (p) { return p.in === 'body' || (p.in === 'formData' && p.type !== 'file'); })
            ? parametersTypeNameTemplate(operationId, 'body')
            : 'unknown';
        var queryParamsType = parameters.some(function (p) { return p.in === 'query'; })
            ? parametersTypeNameTemplate(operationId, 'query')
            : 'unknown';
        var pathParamsType = parameters.some(function (p) { return p.in === 'path'; })
            ? parametersTypeNameTemplate(operationId, 'path')
            : 'unknown';
        var methodName = initLower(operationId);
        var typeName = initUpper(operationId) + "RequestHandler";
        log === null || log === void 0 ? void 0 : log(progress(typeName));
        return {
            methodName: methodName,
            typeName: typeName,
            pathParamsType: pathParamsType,
            responseBodyType: responseBodyType,
            bodyParamType: bodyParamType,
            queryParamsType: queryParamsType,
            statusCodes: statusCodes
        };
    });
    return requestHandlersTemplate(handlers);
};

var fromV3$4 = function (document, log) {
    var handlers = mapOperations$1(document).map(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters, responses = _a.responses, requestBody = _a.requestBody;
        var responseBodyType = responses
            .compactMap(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            if (!response.content ||
                !('application/json' in response.content)) {
                return undefined;
            }
            return responseBodyTypeNameTemplate(operationId, statusCode);
        })
            .join(' | ') || 'unknown';
        var statusCodes = responses
            .map(function (_a) {
            var statusCode = _a.statusCode;
            return statusCode === 'default' ? 'number' : statusCode;
        })
            .join(' | ');
        var bodyParamType = requestBody && Object.keys(requestBody.content).length > 0
            ? parametersTypeNameTemplate(operationId, 'body')
            : 'unknown';
        var queryParamsType = parameters.some(function (p) { return p.in === 'query'; })
            ? parametersTypeNameTemplate(operationId, 'query')
            : 'unknown';
        var pathParamsType = parameters.some(function (p) { return p.in === 'path'; })
            ? parametersTypeNameTemplate(operationId, 'path')
            : 'unknown';
        var methodName = initLower(operationId);
        var typeName = initUpper(operationId) + "RequestHandler";
        log === null || log === void 0 ? void 0 : log(progress(typeName));
        return {
            methodName: methodName,
            typeName: typeName,
            pathParamsType: pathParamsType,
            responseBodyType: responseBodyType,
            bodyParamType: bodyParamType,
            queryParamsType: queryParamsType,
            statusCodes: statusCodes
        };
    });
    return requestHandlersTemplate(handlers);
};

var generateRequestHandlersTypes = function (openApiDocument, log) {
    return 'swagger' in openApiDocument
        ? fromV2$4(openApiDocument, log)
        : fromV3$4(openApiDocument, log);
};

var fromV2$3 = function (document, serviceName, log) {
    var routes = mapOperations(document).map(function (_a) {
        var method = _a.method, path = _a.path, parameters = _a.parameters, operationId = _a.operationId;
        log === null || log === void 0 ? void 0 : log(progress("Adding " + method + " " + path));
        return {
            httpMethod: method,
            openApiPath: path,
            handlerInstanceName: initLower(operationId),
            validatorInstanceName: parameters.any()
                ? operationId + "Validator"
                : undefined
        };
    });
    return routerTemplate({
        serviceName: serviceName,
        routes: routes
    });
};

var fromV3$3 = function (document, serviceName, log) {
    var routes = mapOperations$1(document).map(function (_a) {
        var method = _a.method, path = _a.path, parameters = _a.parameters, operationId = _a.operationId, requestBody = _a.requestBody;
        log === null || log === void 0 ? void 0 : log(progress("Adding " + method + " " + path));
        return {
            httpMethod: method,
            openApiPath: path,
            handlerInstanceName: initLower(operationId),
            validatorInstanceName: parameters.any() || requestBody
                ? operationId + "Validator"
                : undefined
        };
    });
    return routerTemplate({
        serviceName: serviceName,
        routes: routes
    });
};

var generateRouter = function (openApiDocument, serviceName, log) {
    return 'swagger' in openApiDocument
        ? fromV2$3(openApiDocument, serviceName, log)
        : fromV3$3(openApiDocument, serviceName, log);
};

var PERMITTED_TYPES$1 = ['path', 'query', 'header', 'cookie'];
var convertFormDataToSchemaObject = function (formData) {
    if (!formData || !formData.schema)
        return undefined;
    return tslib.__assign({ type: 'object' }, formData.schema);
};
var fromV3$2 = function (document, log) {
    var _a, _b;
    var definitions = {};
    mapOperations$1(document).forEach(function (_a) {
        var _b;
        var operationId = _a.operationId, parameters = _a.parameters, responses = _a.responses, requestBody = _a.requestBody;
        if (requestBody) {
            var typeName = parametersTypeNameTemplate(operationId, 'body');
            log === null || log === void 0 ? void 0 : log(progress("adding " + typeName));
            var structuredContent = [
                (_b = requestBody.content['application/json']) === null || _b === void 0 ? void 0 : _b.schema,
                convertFormDataToSchemaObject(requestBody.content['multipart/form-data'])
            ]
                .compact()
                .map(convertSchemaToJsonSchema$1);
            if (structuredContent.length > 0) {
                definitions[typeName] =
                    structuredContent.length > 1
                        ? {
                            anyOf: structuredContent
                        }
                        : structuredContent[0];
            }
        }
        var groups = parameters.groupBy(function (p) { return p.in; });
        groups.forEach(function (_a) {
            var key = _a.key, values = _a.values;
            if (!PERMITTED_TYPES$1.includes(key)) {
                throw new Error("Unsupported parameter in " + key);
            }
            var typeName = parametersTypeNameTemplate(operationId, key);
            var schema = {
                type: 'object',
                properties: {},
                required: []
            };
            values.forEach(function (v) {
                var _a;
                // eslint-disable-next-line @typescript-eslint/naming-convention
                v.in; var name = v.name, required = v.required, rest = tslib.__rest(v, ["in", "name", "required"]);
                schema.properties[name] = (_a = rest.schema) !== null && _a !== void 0 ? _a : { type: 'string' };
                if (required) {
                    schema.required = tslib.__spreadArray(tslib.__spreadArray([], tslib.__read(schema.required)), [name]);
                }
            });
            log === null || log === void 0 ? void 0 : log(progress("adding " + typeName));
            definitions[typeName] = convertSchemaToJsonSchema$1(schema);
        });
        responses.forEach(function (_a) {
            var _b;
            var statusCode = _a.statusCode, response = _a.response;
            if (!((_b = response.content) === null || _b === void 0 ? void 0 : _b['application/json'])) {
                return;
            }
            var schema = response.content['application/json'].schema;
            if (!schema) {
                return;
            }
            var key = responseBodyTypeNameTemplate(operationId, statusCode);
            log === null || log === void 0 ? void 0 : log(progress("adding " + key));
            definitions[key] = convertSchemaToJsonSchema$1(schema);
        });
    });
    Object.keys((_b = (_a = document.components) === null || _a === void 0 ? void 0 : _a.schemas) !== null && _b !== void 0 ? _b : {}).forEach(function (key) {
        var _a, _b;
        var value = (_b = (_a = document.components) === null || _a === void 0 ? void 0 : _a.schemas) === null || _b === void 0 ? void 0 : _b[key];
        if (value) {
            log === null || log === void 0 ? void 0 : log(progress("adding " + key));
            definitions[key] = convertSchemaToJsonSchema$1(value);
        }
    });
    var out = {
        $schema: 'http://json-schema.org/draft-07/schema#',
        definitions: definitions
    };
    return out;
};

var PERMITTED_TYPES = ['body', 'path', 'query', 'header', 'formData'];
var fromV2$2 = function (document, log) {
    var _a;
    var definitions = {};
    mapOperations(document).forEach(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters, responses = _a.responses;
        var groups = parameters.groupBy(function (p) { return p.in; });
        var keys = groups.map(function (g) { return g.key; });
        if (keys.includes('body') && keys.includes('formData')) {
            throw new Error("Can not have both formData and body parameters");
        }
        groups.forEach(function (_a) {
            var key = _a.key, values = _a.values;
            if (!PERMITTED_TYPES.includes(key)) {
                throw new Error("Unsupported parameter in " + key);
            }
            if (key === 'body') {
                if (values.length > 1) {
                    throw new Error('Only a single body parameter per operation is supported');
                }
                var schema_1 = values[0].schema;
                var typeName_1 = parametersTypeNameTemplate(operationId, 'body');
                log === null || log === void 0 ? void 0 : log(progress("adding " + typeName_1));
                definitions[typeName_1] = convertSchemaToJsonSchema(schema_1);
                return;
            }
            var typeName = parametersTypeNameTemplate(operationId, key);
            var schema = {
                type: 'object',
                properties: {},
                required: []
            };
            values.forEach(function (v) {
                // eslint-disable-next-line @typescript-eslint/naming-convention
                v.in; var name = v.name, required = v.required, rest = tslib.__rest(v, ["in", "name", "required"]);
                if (rest.type === 'file') {
                    return;
                }
                schema.properties[name] = rest;
                if (required) {
                    schema.required = tslib.__spreadArray(tslib.__spreadArray([], tslib.__read(schema.required)), [name]);
                }
            });
            log === null || log === void 0 ? void 0 : log(progress("adding " + typeName));
            definitions[typeName] = convertSchemaToJsonSchema(schema);
        });
        responses.forEach(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var schema = response.schema;
            if (!schema) {
                return;
            }
            if (schema.type === 'file') {
                return;
            }
            var key = responseBodyTypeNameTemplate(operationId, statusCode);
            log === null || log === void 0 ? void 0 : log(progress("adding " + key));
            definitions[key] = convertSchemaToJsonSchema(schema);
        });
    });
    Object.keys((_a = document.definitions) !== null && _a !== void 0 ? _a : {}).forEach(function (key) {
        var _a;
        var value = (_a = document.definitions) === null || _a === void 0 ? void 0 : _a[key];
        if (value) {
            log === null || log === void 0 ? void 0 : log(progress("adding " + key));
            definitions[key] = convertSchemaToJsonSchema(value);
        }
    });
    var out = {
        $schema: 'http://json-schema.org/draft-07/schema#',
        definitions: definitions
    };
    return out;
};

var generateJsonSchema = function (openApiDocument, log) {
    log === null || log === void 0 ? void 0 : log(progress("Generating JSON schema from Open API Document"));
    var schema = 'swagger' in openApiDocument
        ? fromV2$2(openApiDocument, log)
        : fromV3$2(openApiDocument, log);
    return schema;
};

var expressRequestPropertyName = function (type) {
    switch (type) {
        case 'header':
            return 'headers';
        case 'path':
            return 'params';
        case 'formData':
            return 'body';
        default:
            return type;
    }
};

/* eslint-disable no-template-curly-in-string */
var fromV2$1 = function (openApiDocument, log) {
    log === null || log === void 0 ? void 0 : log(progress('Generating request body validators'));
    return mapOperations(openApiDocument)
        .compactMap(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters;
        if (parameters.none()) {
            return undefined;
        }
        var parameterTypes = parameters
            .map(function (p) { return p.in; })
            .distinct()
            .map(function (p) { return ({
            typeName: parametersTypeNameTemplate(operationId, p),
            requestProperty: expressRequestPropertyName(p)
        }); });
        log === null || log === void 0 ? void 0 : log(progress("Generating " + operationId + "Validator"));
        return requestValidatorTemplate({
            functionName: requestValidatorFunctionNameTemplate(operationId),
            parameterTypes: parameterTypes
        });
    })
        .join('\n\n');
};

/* eslint-disable no-template-curly-in-string */
var fromV3$1 = function (openApiDocument, log) {
    log === null || log === void 0 ? void 0 : log(progress('Generating request body validators'));
    return mapOperations$1(openApiDocument)
        .compactMap(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters, requestBody = _a.requestBody;
        var parameterTypes = parameters
            .map(function (p) { return p.in; })
            .distinct()
            .concat(requestBody && Object.keys(requestBody.content).length > 0
            ? ['body']
            : [])
            .map(function (p) { return ({
            typeName: parametersTypeNameTemplate(operationId, p),
            requestProperty: expressRequestPropertyName(p)
        }); });
        if (parameterTypes.none()) {
            return undefined;
        }
        log === null || log === void 0 ? void 0 : log(progress("Generating " + operationId + "Validator"));
        return requestValidatorTemplate({
            functionName: requestValidatorFunctionNameTemplate(operationId),
            parameterTypes: parameterTypes
        });
    })
        .join('\n\n');
};

var generateRequestValidators = function (openApiDocument, log) {
    return 'swagger' in openApiDocument
        ? fromV2$1(openApiDocument, log)
        : fromV3$1(openApiDocument, log);
};

var generateServer = function (_a) {
    var logger = _a.logger, openApiDocument = _a.openApiDocument, serviceName = _a.serviceName;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating server');
    var jsonSchema = generateJsonSchema(openApiDocument, log);
    var types = generateServerTypes(openApiDocument, jsonSchema, log);
    var handlers = generateRequestHandlersTypes(openApiDocument, log);
    var validators = generateRequestValidators(openApiDocument, log);
    var router = generateRouter(openApiDocument, serviceName, log);
    var code = "\n    import Ajv from 'ajv';\n    import { Request, Router } from 'express';\n    import { RequestHandler } from '@laurence79/express-async-request-handler';\n\n    export type ValidationOptions = {\n      logger?: (req: Request) => (message: string, data: Record<string, unknown>) => void;\n    };\n\n    " + types + "\n\n    " + handlers + "\n\n    const ajv = new Ajv({ strict: false });\n    ajv.addSchema(" + JSON.stringify(jsonSchema) + ");\n\n    " + validators + "\n\n    " + router + "\n  ";
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var fromV2 = function (document, log) {
    return mapOperations(document).map(function (_a) {
        var method = _a.method, path = _a.path, parameters = _a.parameters, operationId = _a.operationId, responses = _a.responses;
        log === null || log === void 0 ? void 0 : log(progress("Adding " + method + " " + path));
        var bodyArg = parameters.first(function (p) { return p.in === 'body' || (p.in === 'formData' && p.type !== 'file'); });
        var pathArgs = parameters.filter(function (p) { return p.in === 'path'; });
        var queryArgs = parameters.filter(function (p) { return p.in === 'query'; });
        var headerArgs = parameters.filter(function (p) { return p.in === 'header'; });
        var headersType = headerArgs.any()
            ? parametersTypeNameTemplate(operationId, 'header')
            : 'unknown';
        var pathParamsType = pathArgs.any()
            ? parametersTypeNameTemplate(operationId, 'path')
            : 'unknown';
        var queryParamsType = queryArgs.any()
            ? parametersTypeNameTemplate(operationId, 'query')
            : 'unknown';
        var requestBodyType = bodyArg
            ? parametersTypeNameTemplate(operationId, 'body')
            : 'unknown';
        var responseStubType = responses
            .map(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var statusCodeType = statusCode === 'default' ? 'number' : statusCode;
            var responseBodyType = (function () {
                var schema = response.schema;
                if (!schema) {
                    return 'undefined';
                }
                return responseBodyTypeNameTemplate(operationId, statusCode);
            })();
            return "ResponseStub<" + statusCodeType + ", " + responseBodyType + ">";
        })
            .distinct()
            .join(' | ') || 'ResponseStub';
        var methodStubType = "MethodStub<\n        " + headersType + ",\n        " + pathParamsType + ",\n        " + queryParamsType + ",\n        " + requestBodyType + ",\n        " + responseStubType + "\n      >";
        var requestHandlerPath = path.replace(/\{(?:.*?)\}/g, function (x) { return ":" + x.substr(1, x.length - 2); });
        return {
            operationId: operationId,
            requestHandlerPath: requestHandlerPath,
            method: method,
            methodStubType: methodStubType
        };
    });
};

var fromV3 = function (document, log) {
    return mapOperations$1(document).map(function (_a) {
        var method = _a.method, path = _a.path, parameters = _a.parameters, operationId = _a.operationId, responses = _a.responses, requestBody = _a.requestBody;
        log === null || log === void 0 ? void 0 : log(progress("Adding " + method + " " + path));
        var pathArgs = parameters.filter(function (p) { return p.in === 'path'; });
        var queryArgs = parameters.filter(function (p) { return p.in === 'query'; });
        var headerArgs = parameters.filter(function (p) { return p.in === 'header'; });
        var headersType = headerArgs.any()
            ? parametersTypeNameTemplate(operationId, 'header')
            : 'unknown';
        var pathParamsType = pathArgs.any()
            ? parametersTypeNameTemplate(operationId, 'path')
            : 'unknown';
        var queryParamsType = queryArgs.any()
            ? parametersTypeNameTemplate(operationId, 'query')
            : 'unknown';
        var requestBodyType = requestBody
            ? parametersTypeNameTemplate(operationId, 'body')
            : 'unknown';
        var responseStubType = responses
            .map(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var statusCodeType = statusCode === 'default' ? 'number' : statusCode;
            var responseBodyType = (function () {
                if (!response.content ||
                    Object.keys(response.content).length === 0) {
                    return 'undefined';
                }
                return responseBodyTypeNameTemplate(operationId, statusCode);
            })();
            return "ResponseStub<" + statusCodeType + ", " + responseBodyType + ">";
        })
            .distinct()
            .join(' | ') || 'unknown';
        var methodStubType = "MethodStub<\n        " + headersType + ",\n        " + pathParamsType + ",\n        " + queryParamsType + ",\n        " + requestBodyType + ",\n        " + responseStubType + "\n      >";
        var requestHandlerPath = path.replace(/\{(?:.*?)\}/g, function (x) { return ":" + x.substr(1, x.length - 2); });
        return {
            operationId: operationId,
            requestHandlerPath: requestHandlerPath,
            method: method,
            methodStubType: methodStubType
        };
    });
};

var generateStubs = function (_a) {
    var logger = _a.logger, openApiDocument = _a.openApiDocument, serviceName = _a.serviceName;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating server stubs');
    var jsonSchema = generateJsonSchema(openApiDocument, log);
    var types = generateServerTypes(openApiDocument, jsonSchema, log);
    var data = 'swagger' in openApiDocument
        ? fromV2(openApiDocument, log)
        : fromV3(openApiDocument, log);
    var code = "\n    import { RequestHandler } from '@laurence79/express-async-request-handler';\n    import { Router } from 'express';\n\n    " + types + "\n      \n    type RequestLog<THeaders, TParams, TQuery, TBody> = {\n      headers: THeaders;\n      params: TParams;\n      query: TQuery;\n      body: TBody;\n    };\n\n    type ResponseStub<TStatus extends number = number, TBody = unknown> = {\n      statusCode: TStatus;\n      body: TBody;\n    };\n\n    type MethodStub<\n      THeaders,\n      TParams,\n      TQuery,\n      TRequestBody,\n      TResponseStub extends ResponseStub\n    > = {\n      requestHandler: RequestHandler<\n        TParams,\n        TResponseStub['body'],\n        TRequestBody,\n        TQuery,\n        Record<string, unknown>,\n        TResponseStub['statusCode']\n      >;\n      respondWith(data: TResponseStub): void;\n      callLog(): RequestLog<THeaders, TParams, TQuery, TRequestBody>[];\n    };\n\n    const methodStub = <\n      THeaders,\n      TParams,\n      TQuery,\n      TRequestBody,\n      TResponseStub extends ResponseStub\n    >(): MethodStub<THeaders, TParams, TQuery, TRequestBody, TResponseStub> => {\n      const log: RequestLog<THeaders, TParams, TQuery, TRequestBody>[] = [];\n\n      let responseStub: TResponseStub | null = null;\n\n      return {\n        requestHandler: (req, res) => {\n          const { headers, params, query, body } = req;\n\n          log.push({\n            headers: (headers as unknown) as THeaders,\n            params,\n            query,\n            body\n          });\n\n          if (!responseStub) {\n            throw new Error('Method not stubbed');\n          }\n\n          res.status(responseStub.statusCode).send(responseStub.body);\n        },\n\n        respondWith(data) {\n          responseStub = data;\n        },\n\n        callLog() {\n          return log;\n        }\n      };\n    };\n\n    type MethodStubs = {\n      " + data
        .map(function (_a) {
        var operationId = _a.operationId, methodStubType = _a.methodStubType;
        return "readonly " + operationId + ": " + methodStubType + ";";
    })
        .join('\n') + "\n    };\n\n    export type " + serviceName + "ServiceStub = {\n      readonly middleware: RequestHandler;\n      readonly reset: () => void;\n      readonly stubs: () => MethodStubs;\n    };\n\n    export const " + initLower(serviceName) + "ServiceStub \n    = (): " + serviceName + "ServiceStub => {\n      let methodStubs: MethodStubs;\n      let currentRouter: RequestHandler;\n\n      const reset = () => {\n        methodStubs = {\n          " + data
        .map(function (_a) {
        var operationId = _a.operationId;
        return operationId + ": methodStub()";
    })
        .join(',\n') + "\n        };\n\n        currentRouter = Router()\n          " + data
        .map(function (_a) {
        var method = _a.method, operationId = _a.operationId, requestHandlerPath = _a.requestHandlerPath;
        return "." + method + "('" + requestHandlerPath + "', methodStubs." + operationId + ".requestHandler)";
    })
        .join('\n') + "\n      };\n\n      reset();\n\n      return {\n        reset,\n        middleware: (...args: Parameters<RequestHandler>) => currentRouter(...args),\n        stubs: () => methodStubs\n      };\n    };\n  ";
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

/* eslint-disable @typescript-eslint/restrict-template-expressions */
function assertNever(x) {
    throw new Error("Unexpected object: " + x);
}

var loadOpenApiDocument = function (_a) {
    var openApiDocumentFilename = _a.openApiDocumentFilename, logger = _a.logger;
    var log = logger === null || logger === void 0 ? void 0 : logger.create("Loading swagger file " + openApiDocumentFilename);
    if (!fs__default['default'].existsSync(openApiDocumentFilename)) {
        throw new Error("File " + openApiDocumentFilename + " does not exist");
    }
    var extension = path__default['default'].extname(openApiDocumentFilename);
    var data = fs__default['default'].readFileSync(openApiDocumentFilename, 'UTF-8');
    var doc = (function () {
        if (['.yaml', '.yml'].includes(extension)) {
            return yaml__default['default'].parse(data);
        }
        if (extension === '.json') {
            return JSON.parse(data);
        }
        throw new Error("Unsupported extension " + extension);
    })();
    log === null || log === void 0 ? void 0 : log(success());
    return doc;
};

var prettierStyle = {
    arrowParens: 'avoid',
    semi: true,
    trailingComma: 'none',
    singleQuote: true,
    printWidth: 80
};
var writeFile = function (_a) {
    var content = _a.content, logger = _a.logger, filename = _a.filename;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Writing file');
    log === null || log === void 0 ? void 0 : log(progress('Creating output directory'));
    fs__default['default'].mkdirSync(path__default['default'].dirname(filename), { recursive: true });
    log === null || log === void 0 ? void 0 : log(progress('Formatting content'));
    var extension = path__default['default'].extname(filename);
    var parser = (function () {
        switch (extension) {
            case '.ts':
                return 'typescript';
            case '.json':
                return 'json';
            default:
                return undefined;
        }
    })();
    var formattedContent = (function () {
        try {
            return prettier__default['default'].format(content, tslib.__assign(tslib.__assign({}, prettierStyle), { parser: parser }));
        }
        catch (e) {
            // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
            log === null || log === void 0 ? void 0 : log(warn(e.message));
            return content;
        }
    })();
    fs__default['default'].writeFileSync(filename, formattedContent);
    log === null || log === void 0 ? void 0 : log(success());
};

var defaultOptions = {
    output: 'CLIENT'
};
var generateCode = function (inputOptions) {
    var _a;
    var options = tslib.__assign(tslib.__assign({}, defaultOptions), inputOptions);
    var openApiDocument = loadOpenApiDocument(options);
    var serviceName = (_a = inputOptions.serviceName) !== null && _a !== void 0 ? _a : serviceNameTemplate(openApiDocument.info.title);
    var filename = (function () {
        var file = "" + serviceName + initUpper(options.output.toLowerCase()) + ".generated.ts";
        if (!options.outputFilename) {
            return file;
        }
        if (path__default['default'].extname(options.outputFilename)) {
            return options.outputFilename;
        }
        return path__default['default'].join(options.outputFilename, file);
    })();
    var renderDeps = tslib.__assign(tslib.__assign({}, options), { openApiDocument: openApiDocument,
        serviceName: serviceName,
        filename: filename });
    var method = (function () {
        switch (options.output) {
            case 'CLIENT':
                return generateClient;
            case 'SERVER':
                return generateServer;
            case 'STUBS':
                return generateStubs;
            default:
                return assertNever(options.output);
        }
    })();
    writeFile(tslib.__assign(tslib.__assign({}, renderDeps), { content: method(renderDeps) }));
};

var packageJsonContent = fs__default['default'].readFileSync(__dirname + "/../package.json", 'utf-8');
var packageJson = JSON.parse(packageJsonContent);
var version = packageJson.version;
var cmd = new commander__default['default'].Command('generate')
    .description('Generates typescript code from the spec.')
    .arguments('<spec-filename>')
    .option('--template <template>', 'the type to output. client, server or stubs. Default client')
    .option('--filename <filename>', 'the file to output. Defaults to <service-name><output-type>.generated.ts')
    .option('--service-name <name>', 'the name of the service. Defaults to info.title from the open api document')
    .action(function (specFilename, options) {
    var _a;
    var logger = processLogger();
    if (options.template &&
        !['client', 'server', 'stubs'].includes(options.template.toLowerCase())) {
        commander__default['default'].help({ error: true });
    }
    var output = (options.template
        ? options.template.toUpperCase()
        : 'CLIENT');
    try {
        generateCode({
            openApiDocumentFilename: specFilename,
            output: output,
            outputFilename: options.filename,
            serviceName: options.serviceName,
            logger: logger
        });
        logger.log(chalk__default['default'].green('\nCompleted successfully!\n\n'));
        process.exit(0);
    }
    catch (e) {
        logger.error("\nFAILED. " + (e instanceof Error
            ? e.message + "\n" + ((_a = e.stack) !== null && _a !== void 0 ? _a : '')
            : 'Unknown error') + "\n\n");
        process.exit(1);
    }
});
commander__default['default'].program.version(version).addCommand(cmd).parse();

exports.generateCode = generateCode;
